//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_frame.cpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#include "LaserBoy_space.hpp"

//############################################################################
LaserBoy_frame::LaserBoy_frame(LaserBoy_space* ps,
                               LaserBoy_vertex(*F)(int, int),
                               int vertices_per_frame,
                               int frame_index
                              )
               : LaserBoy_segment(ps)
{
    is_selected            = false;
    is_unique              = true;
    palette_index          = LASERBOY_LASERBOY_HUES;
    error                  = LASERBOY_OK;
    format                 = LASERBOY_3D_VECTOR;
    LaserBoy_vertex vertex = F(0, frame_index);
    vertex.blank();
    reserve(vertices_per_frame + 2);
    push_back(vertex);
    for(int p = 0; p < vertices_per_frame; p++)
        push_back(F(p, frame_index));
    push_back(vertex);
    quantity = (u_short)size();
}

//############################################################################
bool LaserBoy_frame::save_as_ild(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(palette_index != LASERBOY_ILDA_DEFAULT)
          p_space->to_ofstream_ild(out);
        else
            to_ofstream_ild(out);
        out.close();
    }
    else
        return false;
    return true;
}

//############################################################################
void LaserBoy_frame::optimize()
{
    minimize                ();
    add_dwell_vertices      ();
    add_lit_span_vertices   ();
    add_blank_span_vertices ();
    quantity = (u_short)size();
    return;
}

//############################################################################
int LaserBoy_frame::from_ifstream_format_3(ifstream& in, const LaserBoy_header& header)
{
    int error = LASERBOY_OK;
    *this     = header;
    clear();
    //------------------------------------------------------------------------
    if(quantity)
    {
        int              i;
        LaserBoy_color   rgb;
        LaserBoy_vertex  vertex;
        LaserBoy_header  next_header;
        //--------------------------------------------------------------------
        reserve(quantity);
        for(i = 0; i < (int)quantity; i++)
        {
            error |= rgb.from_ifstream_ild(in);
            if(!error)
                push_back((LaserBoy_vertex)rgb);
            else
                return error;
        }
        //--------------------------------------------------------------------
        next_header.from_ifstream_ild(in);

        if(quantity != next_header.quantity)
            return LASERBOY_COLOR_VERTEX_MISSMATCH;

        *this = next_header;

        for(i = 0; i < (int)quantity; i++)
        {
            error |= vertex.from_ifstream_ild(in, next_header.format);
            if(!error)
            {
                at(i).x = vertex.x;
                at(i).y = vertex.y;
                at(i).z = vertex.z;
                at(i).k = vertex.k;
                at(i).c = vertex.c;
            }
            else
                return error;
        }
        //--------------------------------------------------------------------
        front().blank();
        //--------------------------------------------------------------------
        palette_index = LASERBOY_TRUE_COLOR;
        sync_rgb_and_palette();
        //--------------------------------------------------------------------
    }
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int LaserBoy_frame::from_ifstream_ild(ifstream& in, const LaserBoy_header& header)
{
    int              i,
                     error = LASERBOY_OK;
    LaserBoy_vertex  vertex;
    //------------------------------------------------------------------------
    *this = header;
    if(    format == LASERBOY_3D_COLOR_VECTOR
        || format == LASERBOY_2D_COLOR_VECTOR
      )
    {
        palette_index = LASERBOY_TRUE_COLOR;
    }
    //------------------------------------------------------------------------
    clear();
    if(quantity)
    {
        reserve(quantity);
        for(i = 0; i < (int)quantity; i++)
        {
            error |= vertex.from_ifstream_ild(in, format);
            if(!error)
                push_back(vertex);
            else
                return error;
        }
        //--------------------------------------------------------------------
        front().blank();
    }
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int LaserBoy_frame::from_ifstream_dxf(ifstream& in)
{
    int                   i             ;
    double                real_size  = 0,
                          real_scale = 1;
    LaserBoy_3D_double    real_min      ,
                          real_max      ,
                          real_offset   ;
    LaserBoy_vertex       vertex        ;
    LaserBoy_real_segment real_vertices(p_space);
    //------------------------------------------------------------------------
    format        = LASERBOY_3D_VECTOR ;
    palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    real_vertices.from_ifstream_dxf(in);
    //------------------------------------------------------------------------
    if(real_vertices.size() > 2) // more that the default original vector
    {
        if(p_space->auto_scale_dxf)
        {
            real_min =  DBL_MAX;
            real_max = -DBL_MAX;
            for(i = 2; i < (int)real_vertices.size(); i++)
            {
                if(real_vertices[i].x > real_max.x)    real_max.x = real_vertices[i].x;
                if(real_vertices[i].x < real_min.x)    real_min.x = real_vertices[i].x;
                if(real_vertices[i].y > real_max.y)    real_max.y = real_vertices[i].y;
                if(real_vertices[i].y < real_min.y)    real_min.y = real_vertices[i].y;
                if(real_vertices[i].z > real_max.z)    real_max.z = real_vertices[i].z;
                if(real_vertices[i].z < real_min.z)    real_min.z = real_vertices[i].z;
            }
            if(p_space->maintain_dxf_origin)
            {
                if(fabs(real_max.x) > real_size)    real_size = fabs(real_max.x);
                if(fabs(real_min.x) > real_size)    real_size = fabs(real_min.x);
                if(fabs(real_max.y) > real_size)    real_size = fabs(real_max.y);
                if(fabs(real_min.y) > real_size)    real_size = fabs(real_min.y);
                if(fabs(real_max.z) > real_size)    real_size = fabs(real_max.z);
                if(fabs(real_min.z) > real_size)    real_size = fabs(real_min.z);
                real_size *= 2;
            }
            else // find the new center of the universe
            {
                real_offset.x = (real_max.x - ((real_max.x - real_min.x) / 2));
                real_offset.y = (real_max.y - ((real_max.y - real_min.y) / 2));
                real_offset.z = (real_max.z - ((real_max.z - real_min.z) / 2));

                if(fabs(real_max.x - real_min.x) > real_size)
                    real_size = fabs(real_max.x - real_min.x);

                if(fabs(real_max.y - real_min.y) > real_size)
                    real_size = fabs(real_max.y - real_min.y);

                if(fabs(real_max.z - real_min.z) > real_size)
                    real_size = fabs(real_max.z - real_min.z);
            }
            real_scale = 65000.0 / real_size;
        } // end if(auto_scale_dxf)
        //--------------------------------------------------------------------
        reserve(real_vertices.size() - 2);
        for(i = 2; i < (int)real_vertices.size(); i++)
        {
            vertex.x = (short)((real_vertices[i].x - real_offset.x) * real_scale);
            vertex.y = (short)((real_vertices[i].y - real_offset.y) * real_scale);
            vertex.z = (short)((real_vertices[i].z - real_offset.z) * real_scale);
            vertex.r = real_vertices[i].r;
            vertex.g = real_vertices[i].g;
            vertex.b = real_vertices[i].b;
            if(real_vertices[i].is_blank())
                vertex.blank();
            else
                vertex.unblank();
            vertex.set_unique();
            push_back(vertex);
        }
        //set_rgb_from_palette();
    }
    else
    {
        reserve(2);
        push_back(vertex);
        push_back(vertex);
    }
    //------------------------------------------------------------------------
    quantity = (u_short)number_of_vertices();
    is_2D(); // if so change format to LASERBOY_2D_VECTOR
    return LASERBOY_OK;
}

//############################################################################
int LaserBoy_frame::move_selection(LaserBoy_3D_double d)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_frame   frame(*this);
        LaserBoy_3D_double    f;
        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(f + d, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( f + d,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}
//############################################################################
int LaserBoy_frame::scale_selection(LaserBoy_3D_double m)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(((f - center) * m) + center, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( ((f - center) * m) + center,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::scale_selection_on_coordinates(LaserBoy_3D_double m, LaserBoy_3D_double p)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f;

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(((f - p) * m) + p, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( ((f - p) * m) + p,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::scale_selection_on_fulcrum(LaserBoy_3D_double magnitude)
{
    if(number_of_vertices() > 1)
    {
        int  i,
             out_of_bounds = LASERBOY_IN_BOUNDS;

        LaserBoy_3D_double  f;
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(((f - p_space->fulcrum) * magnitude) + p_space->fulcrum, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( ((f - p_space->fulcrum) * magnitude) + p_space->fulcrum,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::rotate_selection(LaserBoy_3D_double a)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(rotate_vertex_on_coordinates(f, center, a), LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( rotate_vertex_on_coordinates(f, center, a),
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::rotate_selection_on_coordinates(LaserBoy_3D_double a, LaserBoy_3D_double p)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f;

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(rotate_vertex_on_coordinates(f, p, a), LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( rotate_vertex_on_coordinates(f, p, a),
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::rotate_selection_on_fulcrum(LaserBoy_3D_double angle)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f;

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(rotate_vertex_on_coordinates(f, p_space->fulcrum, angle), LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( rotate_vertex_on_coordinates(f, p_space->fulcrum, angle),
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
void LaserBoy_frame::next_segment_select()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i + 1, egg, spider);
    return;
}

//############################################################################
void LaserBoy_frame::next_segment_egg()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i + 1, egg, end);
    return;
}

//############################################################################
void LaserBoy_frame::next_segment_spider()
{
    int i, start, end;
    find_segment(spider, start, end, i);
    find_segment(i + 1, start, spider);
    return;
}

//############################################################################
void LaserBoy_frame::previous_segment_select()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i - 1, egg, spider);
    return;
}

//############################################################################
void LaserBoy_frame::previous_segment_egg()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i - 1, egg, end);
    return;
}

//############################################################################
void LaserBoy_frame::previous_segment_spider()
{
    int i, start, end;
    find_segment(spider, start, end, i);
    find_segment(i - 1, start, spider);
    return;
}

//############################################################################
void LaserBoy_frame::recolor_select()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color()
                    && is_index_selected(i)
                  )
                {
                    at(i).c = p_space->selected_palette_color;
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_palette_color].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_palette_color].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_palette_color].b;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color()
                    && is_index_selected(i)
                  )
                {
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_palette_color].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_palette_color].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_palette_color].b;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::recolor_black_in_select()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(is_index_selected(i) && at(i).is_black())
                {
                    at(i).c = p_space->selected_palette_color;
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_palette_color].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_palette_color].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_palette_color].b;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(is_index_selected(i) && at(i).is_black())
                {
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_palette_color].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_palette_color].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_palette_color].b;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::blank_vertices()
{
    if(quantity > 1)
    {
        for(int i = 0; i < number_of_vertices(); i++)
            if(is_index_selected(i))
                at(i).blank();
    }
    return;
}

//############################################################################
void LaserBoy_frame::unblank_vertices()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_vertices(); i++)
            if(is_index_selected(i))
                at(i).unblank();
    }
    return;
}

//############################################################################
void LaserBoy_frame::black_vertices()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_vertices(); i++)
            if(is_index_selected(i) && at(i).is_blank())
            {
                at(i).r = at(i).g = at(i).b = 0x00;
                at(i).c = p_space->palette_picker(palette_index).black;
                at(i).unblank();
            }
    }
    return;
}

//############################################################################
void LaserBoy_frame::unblack_vertices()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_vertices(); i++)
            if(is_index_selected(i) && at(i).color() == 0x00) // blank
                at(i).blank();
    }
    return;
}

//############################################################################
void LaserBoy_frame::blank_equivalent_vectors()
{
    LaserBoy_frame_set  segments(p_space);
    int                 i,
                        j;
    //------------------------------------------------------------------------
    fracture_segments();
    segments = find_segments();
    //------------------------------------------------------------------------
    for(i = 0; i < segments.number_of_frames() - 1; i++)
    {
        for(j = i + 1; j < segments.number_of_frames(); j++)
            if(    segments[j] == segments[i]
                || segments[j] == segments[i].reverse()
              )
                segments[j].blank_all_vertices();
    }
    //------------------------------------------------------------------------
    *this = segments.sum_of_frames();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_frame::minimize()
{
    if(!p_space->allow_lit_black)
        convert_black_to_blank();
    //------------------------------------------------------------------------
    remove_dwell_vertices ();
    reduce_blank_vectors();
    reduce_lit_vectors  ();
    //------------------------------------------------------------------------
    if(p_space->no_equivalent_vectors)
    {
        blank_equivalent_vectors();
        reduce_blank_vectors    ();
    }
    //------------------------------------------------------------------------
    if(p_space->no_dots)
        remove_dots ();
    else
        enhance_dots();
    //------------------------------------------------------------------------
    bond_segments();
    validate     ();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::add_new_vertex(int vertex_index)
{
    LaserBoy_segment::add_new_vertex(vertex_index);
    quantity = (u_short)size();
    if(    quantity > 2
        && vertex_index >= 0
        && vertex_index <  quantity
      )
    {
        if((at(vertex_index) | at(vertex_index + 1)) > p_space->insignificant_distance)
            spider = vertex_index + 1;
        else
            spider = vertex_index;
    }
    else
        egg = spider = vertex_index = 0;
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::break_segment()
{
    LaserBoy_segment::break_segment(spider);
    quantity = (u_short)size();
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::connect_the_dots()
{
    LaserBoy_segment::connect_the_dots(egg, spider);
    quantity = (u_short)size();
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::to_dots()
{
    int             i;
    LaserBoy_frame  frame(p_space, palette_index);
    //--------------------------------------------------------------------
    frame.reserve(number_of_vertices() + size_of_selection());
    //--------------------------------------------------------------------
    for(i = 0; i < number_of_vertices(); i++)
    {
        if(    at(i).is_lit()
            && at_index_has_magnitude(i)
          )
        {
            frame += at(i);
            frame.back().r = 0;
            frame.back().g = 0;
            frame.back().b = 0;
            frame.back().c = p_space->palette_picker(palette_index).black;
        }
        frame += at(i);
    }
    //--------------------------------------------------------------------
    *this = frame;
    quantity = (u_short)size();
    //--------------------------------------------------------------------
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::selected_to_dots()
{
    int             i,
                    new_spider = spider;
    LaserBoy_frame  frame(p_space, palette_index);
    //--------------------------------------------------------------------
    frame.reserve(number_of_vertices() + size_of_selection());
    //--------------------------------------------------------------------
    frame.push_back(front());
    for(i = 1; i < number_of_vertices(); i++)
    {
        if(    is_index_selected(i)
            && at(i).is_lit()
            && at_index_has_magnitude(i)
          )
        {
            frame += at(i);
            frame.back().r = 0;
            frame.back().g = 0;
            frame.back().b = 0;
            frame.back().c = p_space->palette_picker(palette_index).black;
            new_spider++;
        }
        frame += at(i);
    }
    //--------------------------------------------------------------------
    spider = new_spider;
    *this = frame;
    quantity = (u_short)size();
    //--------------------------------------------------------------------
    return *this;
}

//############################################################################
void LaserBoy_frame::to_ofstream_ild(ofstream& out)
{
    char number[7];
    if(p_space->dumb_file_format)
    {
        sprintf(number, "%06d", identity);
        name = "ES" + string(number);
        LaserBoy_header::to_ofstream_ild(out);
        for(int i = 0; i < (int)quantity; i++)
            at(i).to_ofstream_ild(out, format);
    }
    else
    {
        if(palette_index == LASERBOY_TRUE_COLOR)
        {
            LaserBoy_header  header  = (LaserBoy_header)*this;
            LaserBoy_palette      palette = as_LaserBoy_palette();
            header.identity--;
            sprintf(number, "%06d", header.identity);
            header.format    = LASERBOY_COLOR_TABLE;
            header.name      = "ES" + string(number);
            palette.to_ofstream_ild(out, header);
        }
        else
        {
            sprintf(number, "%06d", identity);
            name = "ES" + string(number);
        }
        LaserBoy_header::to_ofstream_ild(out);
        for(int i = 0; i < (int)quantity; i++)
            at(i).to_ofstream_ild(out, format);
    }
    return;
}

//############################################################################
void LaserBoy_frame::render() const
{
    u_short      i;

    LaserBoy_3D_double _0,
                       _1;

    LaserBoy_color  bit_masked;

    u_int        select_start_color,
                 select_end_color  ,
                 bmp_bg_color = p_space->bmp.make_rgb(p_space->bg_color.r, p_space->bg_color.g, p_space->bg_color.b);

    int          xx,
                 yy,
                 x_off        = p_space->bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * p_space->font_size_factor,
                 y_off        =   LASERBOY_PALETTE_MAX * p_space->font_size_factor
                                + (LASERBOY_STATS_DISPLAY_OFFSET + 24) * p_space->font_size_factor,
                 y_line       = 8 * p_space->font_size_factor,
                 bmp_center   = p_space->bmp.yres / 2;

    double       scale        = p_space->bmp.yres / 65536.0;

    bool         rotate_view  = (p_space->view_angle  != 0.0),
                 offset_view  = (p_space->view_offset != 0.0),
                 scale_view   = (p_space->view_scale  != 1.0);
    //------------------------------------------------------------------------
    if(p_space->show_background_bitmap)
        bmp_copy(&(p_space->bmp), &(p_space->background_bmp));
    else
        bmp_copy(&(p_space->bmp), &(p_space->background));
    //------------------------------------------------------------------------
    if(p_space->show_fixed_bounds)
        bmp_put_rectangle(&(p_space->bmp), 0, 0, p_space->bmp.yres - 1, p_space->bmp.yres - 1, p_space->bmp.gray, 0);
    //------------------------------------------------------------------------
    if(p_space->show_fixed_origin)
    {
        bmp_put_line_fast(&(p_space->bmp), 0, p_space->bmp.yres / 2, p_space->bmp.yres - 1, p_space->bmp.yres / 2, p_space->bmp.gray);
        bmp_put_line_fast(&(p_space->bmp), p_space->bmp.yres / 2, 0, p_space->bmp.yres / 2, p_space->bmp.yres - 1, p_space->bmp.gray);
    }
    //------------------------------------------------------------------------
    if(p_space->show_floating_axis)
    {
        LaserBoy_3D_double  x1(LASERBOY_MIN_SHORT, 0, 0)
                          , x2(LASERBOY_MAX_SHORT, 0, 0)
                          , y1(0, LASERBOY_MIN_SHORT, 0)
                          , y2(0, LASERBOY_MAX_SHORT, 0)
                          , z1(0, 0, LASERBOY_MIN_SHORT)
                          , z2(0, 0, LASERBOY_MAX_SHORT);

        if(rotate_view)
        {
            x1 = rotate_vertex(x1, p_space->view_angle);
            x2 = rotate_vertex(x2, p_space->view_angle);
            y1 = rotate_vertex(y1, p_space->view_angle);
            y2 = rotate_vertex(y2, p_space->view_angle);
            z1 = rotate_vertex(z1, p_space->view_angle);
            z2 = rotate_vertex(z2, p_space->view_angle);
        }
        if(offset_view)
        {
            x1 += p_space->view_offset;
            x2 += p_space->view_offset;
            y1 += p_space->view_offset;
            y2 += p_space->view_offset;
        }
        if(scale_view)
        {
            x1 *= p_space->view_scale;
            x2 *= p_space->view_scale;
            y1 *= p_space->view_scale;
            y2 *= p_space->view_scale;
            z1 *= p_space->view_scale;
            z2 *= p_space->view_scale;
        }
        // X axis
        bmp_put_line_fast(&(p_space->bmp),           (int)(x1.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(x1.y * scale) + bmp_center)
                              ,                      (int)(x2.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(x2.y * scale) + bmp_center)
                              , p_space->bmp.gray
                         );
        // X axis icon
        bmp_put_line_fast(&(p_space->bmp),           (int)((x2.x - 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((x2.y - 500) * scale) + bmp_center)
                              ,                      (int)((x2.x + 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((x2.y + 500) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(255,0,0)
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)((x2.x - 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((x2.y + 500) * scale) + bmp_center)
                              ,                      (int)((x2.x + 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((x2.y - 500) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(255,0,0)
                         );
        // Y axis
        bmp_put_line_fast(&(p_space->bmp),           (int)(y1.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(y1.y * scale) + bmp_center)
                              ,                      (int)(y2.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(y2.y * scale) + bmp_center)
                              , p_space->bmp.gray
                         );
        // Y axis icon
        bmp_put_line_fast(&(p_space->bmp),           (int)((y2.x - 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((y2.y + 500) * scale) + bmp_center)
                              ,                      (int)((y2.x      ) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((y2.y      ) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)((y2.x + 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((y2.y + 500) * scale) + bmp_center)
                              ,                      (int)((y2.x      ) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((y2.y      ) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)((y2.x      ) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((y2.y      ) * scale) + bmp_center)
                              ,                      (int)((y2.x      ) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((y2.y - 500) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,255,0)
                         );
        // Z axis
        bmp_put_line_fast(&(p_space->bmp),           (int)(z1.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(z1.y * scale) + bmp_center)
                              ,                      (int)(z2.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(z2.y * scale) + bmp_center)
                              , p_space->bmp.gray
                         );
        // Z axis icon
        bmp_put_line_fast(&(p_space->bmp),           (int)((z2.x - 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              ,                      (int)((z2.x + 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,0,255)
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)((z2.x + 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              ,                      (int)((z2.x - 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,0,255)
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)((z2.x - 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              ,                      (int)((z2.x + 500) * scale) + bmp_center
                              , p_space->bmp.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,0,255)
                         );
    }
    //------------------------------------------------------------------------
    if(p_space->show_floating_bounds)
    {
        LaserBoy_3D_double  a(LASERBOY_MIN_SHORT, LASERBOY_MIN_SHORT, LASERBOY_MIN_SHORT)
                          , b(LASERBOY_MAX_SHORT, LASERBOY_MIN_SHORT, LASERBOY_MIN_SHORT)
                          , c(LASERBOY_MAX_SHORT, LASERBOY_MIN_SHORT, LASERBOY_MAX_SHORT)
                          , d(LASERBOY_MIN_SHORT, LASERBOY_MIN_SHORT, LASERBOY_MAX_SHORT)
                          , e(LASERBOY_MIN_SHORT, LASERBOY_MAX_SHORT, LASERBOY_MIN_SHORT)
                          , f(LASERBOY_MAX_SHORT, LASERBOY_MAX_SHORT, LASERBOY_MIN_SHORT)
                          , g(LASERBOY_MAX_SHORT, LASERBOY_MAX_SHORT, LASERBOY_MAX_SHORT)
                          , h(LASERBOY_MIN_SHORT, LASERBOY_MAX_SHORT, LASERBOY_MAX_SHORT);

        if(rotate_view)
        {
            a = rotate_vertex(a, p_space->view_angle);
            b = rotate_vertex(b, p_space->view_angle);
            c = rotate_vertex(c, p_space->view_angle);
            d = rotate_vertex(d, p_space->view_angle);
            e = rotate_vertex(e, p_space->view_angle);
            f = rotate_vertex(f, p_space->view_angle);
            g = rotate_vertex(g, p_space->view_angle);
            h = rotate_vertex(h, p_space->view_angle);
        }
        if(offset_view)
        {
            a += p_space->view_offset;
            b += p_space->view_offset;
            c += p_space->view_offset;
            d += p_space->view_offset;
            e += p_space->view_offset;
            f += p_space->view_offset;
            g += p_space->view_offset;
            h += p_space->view_offset;
        }
        if(scale_view)
        {
            a *= p_space->view_scale;
            b *= p_space->view_scale;
            c *= p_space->view_scale;
            d *= p_space->view_scale;
            e *= p_space->view_scale;
            f *= p_space->view_scale;
            g *= p_space->view_scale;
            h *= p_space->view_scale;
        }

        a *= scale;
        b *= scale;
        c *= scale;
        d *= scale;
        e *= scale;
        f *= scale;
        g *= scale;
        h *= scale;

        bmp_put_line_fast(&(p_space->bmp),           (int)a.x + bmp_center
                              , p_space->bmp.yres - ((int)a.y + bmp_center)
                              ,                      (int)b.x + bmp_center
                              , p_space->bmp.yres - ((int)b.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)b.x + bmp_center
                              , p_space->bmp.yres - ((int)b.y + bmp_center)
                              ,                      (int)c.x + bmp_center
                              , p_space->bmp.yres - ((int)c.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)c.x + bmp_center
                              , p_space->bmp.yres - ((int)c.y + bmp_center)
                              ,                      (int)d.x + bmp_center
                              , p_space->bmp.yres - ((int)d.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)d.x + bmp_center
                              , p_space->bmp.yres - ((int)d.y + bmp_center)
                              ,                      (int)a.x + bmp_center
                              , p_space->bmp.yres - ((int)a.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)a.x + bmp_center
                              , p_space->bmp.yres - ((int)a.y + bmp_center)
                              ,                      (int)e.x + bmp_center
                              , p_space->bmp.yres - ((int)e.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)b.x + bmp_center
                              , p_space->bmp.yres - ((int)b.y + bmp_center)
                              ,                      (int)f.x + bmp_center
                              , p_space->bmp.yres - ((int)f.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)c.x + bmp_center
                              , p_space->bmp.yres - ((int)c.y + bmp_center)
                              ,                      (int)g.x + bmp_center
                              , p_space->bmp.yres - ((int)g.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)d.x + bmp_center
                              , p_space->bmp.yres - ((int)d.y + bmp_center)
                              ,                      (int)h.x + bmp_center
                              , p_space->bmp.yres - ((int)h.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)e.x + bmp_center
                              , p_space->bmp.yres - ((int)e.y + bmp_center)
                              ,                      (int)f.x + bmp_center
                              , p_space->bmp.yres - ((int)f.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)f.x + bmp_center
                              , p_space->bmp.yres - ((int)f.y + bmp_center)
                              ,                      (int)g.x + bmp_center
                              , p_space->bmp.yres - ((int)g.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)g.x + bmp_center
                              , p_space->bmp.yres - ((int)g.y + bmp_center)
                              ,                      (int)h.x + bmp_center
                              , p_space->bmp.yres - ((int)h.y + bmp_center)
                              , p_space->bmp.gray
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)h.x + bmp_center
                              , p_space->bmp.yres - ((int)h.y + bmp_center)
                              ,                      (int)e.x + bmp_center
                              , p_space->bmp.yres - ((int)e.y + bmp_center)
                              , p_space->bmp.gray
                         );
    }
    //------------------------------------------------------------------------
    if(p_space->show_fulcrum)
    {
        LaserBoy_3D_double  x1(-5000, 0, 0)
                          , x2( 5000, 0, 0)
                          , y1(0, -5000, 0)
                          , y2(0,  5000, 0)
                          , z1(0, 0, -5000)
                          , z2(0, 0,  5000);

        x1 = x1 += p_space->fulcrum;
        x2 = x2 += p_space->fulcrum;
        y1 = y1 += p_space->fulcrum;
        y2 = y2 += p_space->fulcrum;
        z1 = z1 += p_space->fulcrum;
        z2 = z2 += p_space->fulcrum;

        if(rotate_view)
        {
            x1 = rotate_vertex(x1, p_space->view_angle);
            x2 = rotate_vertex(x2, p_space->view_angle);
            y1 = rotate_vertex(y1, p_space->view_angle);
            y2 = rotate_vertex(y2, p_space->view_angle);
            z1 = rotate_vertex(z1, p_space->view_angle);
            z2 = rotate_vertex(z2, p_space->view_angle);
        }
        if(offset_view)
        {
            x1 += p_space->view_offset;
            x2 += p_space->view_offset;
            y1 += p_space->view_offset;
            y2 += p_space->view_offset;
        }
        if(scale_view)
        {
            x1 *= p_space->view_scale;
            x2 *= p_space->view_scale;
            y1 *= p_space->view_scale;
            y2 *= p_space->view_scale;
        }
        bmp_put_line_fast(&(p_space->bmp),           (int)(x1.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(x1.y * scale) + bmp_center)
                              ,                      (int)(x2.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(x2.y * scale) + bmp_center)
                              , p_space->bmp.make_rgb(255,0,0)
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)(y1.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(y1.y * scale) + bmp_center)
                              ,                      (int)(y2.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(y2.y * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&(p_space->bmp),           (int)(z1.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(z1.y * scale) + bmp_center)
                              ,                      (int)(z2.x * scale) + bmp_center
                              , p_space->bmp.yres - ((int)(z2.y * scale) + bmp_center)
                              , p_space->bmp.make_rgb(0,0,255)
                         );
    }
    //------------------------------------------------------------------------
    // ACTUAL VECTOR ART !!!
    //------------------------------------------------------------------------
    if(quantity > 1)
    {
        //--------------------------------------------------------------------
        if(p_space->show_blanking)
        {
            for(i = 1; i < quantity; i++)
            {
                if(at(i).is_blank())
                {
                    if(rotate_view)
                    {
                        _0 = rotate_vertex(at(i - 1).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                        _1 = rotate_vertex(at(i    ).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                    }
                    else
                    {
                        _0 = at(i - 1).bit_masked_position(p_space->signal_bit_mask);
                        _1 = at(i    ).bit_masked_position(p_space->signal_bit_mask);
                    }
                    if(offset_view)
                    {
                        _0 += p_space->view_offset;
                        _1 += p_space->view_offset;
                    }
                    if(scale_view)
                    {
                        _0 *= p_space->view_scale;
                        _1 *= p_space->view_scale;
                    }
                    xx = (int)(_1.x * scale) + bmp_center;
                    yy = (int)(_1.y * scale) + bmp_center;

                    bmp_put_line_fast(&(p_space->bmp),           (int)(_0.x * scale) + bmp_center
                                          , p_space->bmp.yres - ((int)(_0.y * scale) + bmp_center)
                                          ,                      xx
                                          , p_space->bmp.yres - (yy)
                                          , p_space->bmp.gray
                                     );
                    if(p_space->show_vertices)
                        bmp_put_vertex(&(p_space->bmp), xx - 1
                                           , p_space->bmp.yres - yy
                                           , p_space->bmp.gray
                                      );
                    if(p_space->show_cursors && is_index_selected(i))
                        bmp_put_select(&(p_space->bmp), xx
                                           , p_space->bmp.yres - yy
                                           , p_space->bmp.gray
                                      );
                }
            }
        } // end if(p_space->show_blanking)
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).is_black())
            {
                if(rotate_view)
                {
                    _0 = rotate_vertex(at(i - 1).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                    _1 = rotate_vertex(at(i    ).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                }
                else
                {
                    _0 = at(i - 1).bit_masked_position(p_space->signal_bit_mask);
                    _1 = at(i    ).bit_masked_position(p_space->signal_bit_mask);
                }
                if(offset_view)
                {
                    _0 += p_space->view_offset;
                    _1 += p_space->view_offset;
                }
                if(scale_view)
                {
                    _0 *= p_space->view_scale;
                    _1 *= p_space->view_scale;
                }
                xx = (int)(_1.x * scale) + bmp_center;
                yy = (int)(_1.y * scale) + bmp_center;
                //------------------------------------------------------------
                bmp_put_line_fast(&(p_space->bmp),             (int)(_0.x * scale) + bmp_center
                                      , p_space->bmp.yres - ((int)(_0.y * scale) + bmp_center)
                                      ,             xx
                                      , p_space->bmp.yres - (yy)
                                      , 0x00000000
                                 );
                if(p_space->show_vertices)
                    bmp_put_vertex(&(p_space->bmp), xx
                                       , p_space->bmp.yres - yy
                                       , 0x00000000
                                  );
                //------------------------------------------------------------
                if(p_space->show_cursors && is_index_selected(i))
                    bmp_put_select(&(p_space->bmp), xx
                                       , p_space->bmp.yres - yy
                                       , p_space->bmp.white
                                  );
            }
        }
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).is_color())
            {
                if(rotate_view)
                {
                    _0 = rotate_vertex(at(i - 1).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                    _1 = rotate_vertex(at(i    ).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                }
                else
                {
                    _0 = at(i - 1).bit_masked_position(p_space->signal_bit_mask);
                    _1 = at(i    ).bit_masked_position(p_space->signal_bit_mask);
                }
                if(offset_view)
                {
                    _0 += p_space->view_offset;
                    _1 += p_space->view_offset;
                }
                if(scale_view)
                {
                    _0 *= p_space->view_scale;
                    _1 *= p_space->view_scale;
                }

                xx = (int)(_1.x * scale) + bmp_center;
                yy = (int)(_1.y * scale) + bmp_center;
                bit_masked = at(i).bit_masked_color(p_space->signal_bit_mask);
                //------------------------------------------------------------
                bmp_put_line_fast(&(p_space->bmp),           (int)(_0.x * scale) + bmp_center
                                      , p_space->bmp.yres - ((int)(_0.y * scale) + bmp_center)
                                      ,                      xx
                                      , p_space->bmp.yres - (yy)
                                      , p_space->bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b)
                                 );
                if(p_space->show_vertices)
                    bmp_put_vertex(&(p_space->bmp), xx
                                       , p_space->bmp.yres - yy
                                       , p_space->bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b)
                                  );
                //------------------------------------------------------------
                if(p_space->show_cursors && is_index_selected(i))
                    bmp_put_select(&(p_space->bmp), xx
                                       , p_space->bmp.yres - yy
                                       ,   at(i).is_black()
                                         ? p_space->bmp.white
                                         : p_space->bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b)
                                  );
            }
        }
        //--------------------------------------------------------------------
        // END ACTUAL VECTOR ART !!!
        //--------------------------------------------------------------------
        if(p_space->show_cursors)
        {
            bit_masked = at(egg).bit_masked_color(p_space->signal_bit_mask);
            select_start_color =   at(egg).is_blank()
                                 ? p_space->bmp.gray
                                 :   at(egg).is_black()
                                   ? p_space->bmp.white
                                   : p_space->bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
            //----------------------------------------------------------------
            bit_masked = at(spider).bit_masked_color(p_space->signal_bit_mask);
            select_end_color   =   at(spider).is_blank()
                                 ? p_space->bmp.gray
                                 :   at(spider).is_black()
                                   ? p_space->bmp.white
                                   : p_space->bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
            //----------------------------------------------------------------
            if(rotate_view)
            {
                _0 = rotate_vertex(at(egg   ).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                _1 = rotate_vertex(at(spider).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
            }
            else
            {
                _0 = at(egg   ).bit_masked_position(p_space->signal_bit_mask);
                _1 = at(spider).bit_masked_position(p_space->signal_bit_mask);
            }
            if(offset_view)
            {
                _0 += p_space->view_offset;
                _1 += p_space->view_offset;
            }
            if(scale_view)
            {
                _0 *= p_space->view_scale;
                _1 *= p_space->view_scale;
            }
            //----------------------------------------------------------------
            // start vertex select cursor
            bmp_put_rectangle(&(p_space->bmp),           (int)((_0.x - 1000) * scale) + bmp_center
                                  , p_space->bmp.yres - ((int)((_0.y - 1000) * scale) + bmp_center)
                                  ,                      (int)((_0.x + 1000) * scale) + bmp_center
                                  , p_space->bmp.yres - ((int)((_0.y + 1000) * scale) + bmp_center)
                                  , select_start_color
                                  , 0
                             );
            //----------------------------------------------------------------
            // end vertex select cursor
            bmp_put_line_fast (&(p_space->bmp),           (int)((_1.x       ) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                   ,                      (int)((_1.x       ) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                   , select_end_color
                              );
            bmp_put_line_fast (&(p_space->bmp),           (int)((_1.x + 1000) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y       ) * scale) + bmp_center)
                                   ,                      (int)((_1.x - 1000) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y       ) * scale) + bmp_center)
                                   , select_end_color
                              );
            bmp_put_line_fast (&(p_space->bmp),           (int)((_1.x - 1000) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                   ,                      (int)((_1.x + 1000) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                   , select_end_color
                              );
            bmp_put_line_fast (&(p_space->bmp),           (int)((_1.x + 1000) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                   ,                      (int)((_1.x - 1000) * scale) + bmp_center
                                   , p_space->bmp.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                   , select_end_color
                              );
            //----------------------------------------------------------------
        } // end if(p_space->show_cursors)
        //--------------------------------------------------------------------
    } // end if(quantity > 1)
    return;
}

//############################################################################
bool LaserBoy_frame::save_as_wave(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    //------------------------------------------------------------------------
    if(out.is_open())
    {
        if(quantity > 1)
        {
            int                   i,
                                  frame_scan,
                                  blank_dwell = 0,
                                  total_frame_scans;

            LaserBoy_vertex       vertex;

            LaserBoy_segment      intro  (p_space),
                                  bridge (p_space),
                                  coda   (p_space);

            LaserBoy_frame        copy(*this);

            LaserBoy_wave_header  header(p_space->sample_rate,
                                           LASERBOY_WAVE_POSITIVE
                                         | LASERBOY_WAVE_END_OF_FRAME
                                         | LASERBOY_WAVE_UNIQUE_FRAME
                                         | LASERBOY_WAVE_SIGNAL_MATRIX,
                                         6
                                        );
            //----------------------------------------------------------------
            header.signal_id [0] = LASERBOY_SIGNAL_X_POSITION  ; // default values
            header.signal_id [1] = LASERBOY_SIGNAL_Y_POSITION  ;
            header.signal_id [2] = LASERBOY_SIGNAL_RED_ANALOG  ;
            header.signal_id [3] = LASERBOY_SIGNAL_GREEN_ANALOG;
            header.signal_id [4] = LASERBOY_SIGNAL_BLUE_ANALOG ;
            header.signal_id [5] = LASERBOY_SIGNAL_Z_POSITION  ;

            header.LSB_tag   [0] = LASERBOY_LSB_NOT_USED    ; // default values
            header.LSB_tag   [1] = LASERBOY_LSB_NOT_USED    ;
            header.LSB_tag   [2] = LASERBOY_LSB_END_OF_FRAME;
            header.LSB_tag   [3] = LASERBOY_LSB_UNIQUE_FRAME;
            header.LSB_tag   [4] = LASERBOY_LSB_NOT_USED    ;
            header.LSB_tag   [5] = LASERBOY_LSB_NOT_USED    ;

            if(    p_space->signal_bit_mask[0] != 0
                || p_space->signal_bit_mask[1] != 0
                || p_space->signal_bit_mask[2]  > 7
                || p_space->signal_bit_mask[3]  > 7
                || p_space->signal_bit_mask[4]  > 7
                || p_space->signal_bit_mask[5] != 0
              )
            {
                header.resolution[0] = 16 - p_space->signal_bit_mask[0];
                header.resolution[1] = 16 - p_space->signal_bit_mask[1];
                header.resolution[2] = 16 - p_space->signal_bit_mask[2];
                header.resolution[3] = 16 - p_space->signal_bit_mask[3];
                header.resolution[4] = 16 - p_space->signal_bit_mask[4];
                header.resolution[5] = 16 - p_space->signal_bit_mask[5];
                header.LaserBoy_wave_mode |= LASERBOY_SIGNAL_BIT_RESOLUTION;
            }
            //----------------------------------------------------------------
            copy.rotate_around_origin();
            //----------------------------------------------------------------
            if(copy.is_2D() || p_space->auto_flatten_z)
            {
                header.signal_id [5] = p_space->signal_if_not_z;
                header.resolution[5] = 8;
                copy.flatten_z();
            }
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
            {
                header.LaserBoy_wave_mode |= LASERBOY_WAVE_OPTIMIZED;
                header.parms.lit_dwell_overhang     =        p_space->lit_dwell_overhang;
                header.parms.lit_delta_max          = (float)p_space->lit_delta_max;
                header.parms.blank_delta_max        = (float)p_space->blank_delta_max;
                header.parms.max_dwell_microsec     = (float)p_space->max_dwell_microsec;
                header.parms.insignificant_distance = (float)p_space->insignificant_distance;
                header.parms.insignificant_angle    = (float)p_space->insignificant_angle;
                header.parms.frames_per_second      = (float)p_space->frames_per_second;
                copy.optimize();
            }
            //----------------------------------------------------------------
            total_frame_scans = (int)ceil((   p_space->sample_rate
                                            * p_space->still_frame_duration
                                          ) / copy.quantity
                                         );
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
            {
                blank_dwell = (int)(   p_space->sample_rate
                                     / p_space->max_dwell_microsec
                                   );
                //------------------------------------------------------------
                for(i = 0; i < blank_dwell; i++)
                    intro.push_back(LaserBoy_vertex()); // the origin
                //------------------------------------------------------------
                vertex = copy.first_lit_anchor();
                vertex.blank();
                intro += LaserBoy_segment(p_space, intro[0], vertex); // line from origin to first vertex
                //------------------------------------------------------------
                bridge = LaserBoy_segment(p_space, copy.last_lit_vertex(), vertex); // line from last vertex to first for repete
                bridge.blank_all_vertices();
                //------------------------------------------------------------
                coda = LaserBoy_segment(p_space, vertex, intro[0]); // line from last vertex to origin
                for(i = 0; i < blank_dwell; i++)
                    coda.push_back(LaserBoy_vertex()); // the origin
            } // end if(p_space->auto_optimize)
            //----------------------------------------------------------------
            if(p_space->invert_wave_output)
                header.LaserBoy_wave_mode &= ~LASERBOY_WAVE_POSITIVE;
            //----------------------------------------------------------------
            header.to_ofstream_wave(out); // take up the space!
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
                intro.to_ofstream_wave(out, header, false, false);
            //----------------------------------------------------------------
            ((LaserBoy_segment)copy).to_ofstream_wave(out, header, true, is_unique); // first frame unique
            //----------------------------------------------------------------
            for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
            {
                bridge.to_ofstream_wave(out, header, false, false);
                ((LaserBoy_segment)copy).to_ofstream_wave(out, header, true, false); // the rest are not
            }
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
                coda.to_ofstream_wave(out, header, false, false);
            //----------------------------------------------------------------
            header.to_ofstream_wave(out);
            //----------------------------------------------------------------
        }
        out.close();
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame::to_ofstream_wave(ofstream& out, LaserBoy_wave_header& header)
{
    if(quantity > 1)
    {
        int frame_scan,
            total_frame_scans = 1;
        //--------------------------------------------------------------------
        LaserBoy_segment bridge(p_space, last_lit_vertex(), first_lit_anchor()); // line from last vertex to first for repete
        bridge.blank_all_vertices();
        //--------------------------------------------------------------------
        if(    p_space->auto_optimize
            && p_space->frames_per_second > 0.00
          )
        {
            total_frame_scans = (int)ceil(   (    p_space->sample_rate
                                                / (   quantity
                                                    + bridge.number_of_vertices()
                                                  )
                                              )
                                            / p_space->frames_per_second
                                         );
        } // end if(p_space->auto_optimize && p_space->frames_per_second > 0.00)
        //--------------------------------------------------------------------
        LaserBoy_segment::to_ofstream_wave(out, header, true, true); // first pass through this frame
        //--------------------------------------------------------------------
        if(total_frame_scans > 1)
            for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
            {
                bridge.to_ofstream_wave(out, header, false, false);
                LaserBoy_segment::to_ofstream_wave(out, header, true, false);
            }
        //--------------------------------------------------------------------
        transition_to_next_frame.to_ofstream_wave(out, header, false, false);
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_frame::to_ofstream_code(ofstream& out, const char* indent) const
{
    out << indent
        << "frame.format = "
        <<  (int)format
        <<  ";\n"
        <<  indent
        << "frame.palette_index = "
        <<  palette_index
        <<  ";\n";

    for(int i = 0; i < quantity; i++)
        out << indent
            <<  "frame.push_back(LaserBoy_vertex("
            << setw(6)
            <<  at(i).x
            <<  ", "
            << setw(6)
            <<  at(i).y
            <<  ", "
            << setw(6)
            <<  at(i).z
            <<  ", "
            << setw(3)
            <<  (int)at(i).r
            <<  ", "
            << setw(3)
            <<  (int)at(i).g
            <<  ", "
            << setw(3)
            <<  (int)at(i).b
            <<  ", "
            << setw(3)
            <<  (int)at(i).k
            <<  ", "
            << setw(3)
            <<  (int)at(i).c
            <<  "));\n";

    out << indent << "frame.quantity = (u_short)frame.quantity;\n";
    return;
}

//############################################################################
bool LaserBoy_frame::save_as_code(string object_name, string dir) const
{
    ofstream out((dir + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"

               "LaserBoy_frame "
            << object_name
            << "()\n"
               "{\n"
               "    LaserBoy_frame frame;\n";

        to_ofstream_code(out, "    ");

        out << "    return frame;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
        return true;
    }
    return false;
}

//############################################################################
bool LaserBoy_frame::save_as_dxf(string file) const
{
    ofstream out(file.c_str(), ios::out);
    if(out.is_open())
    {
        int                segment_vector_index,
                           vertex_index;
        LaserBoy_frame_set  segments = find_segments();
        //--------------------------------------------------------------------
        segments.minimize();
        //--------------------------------------------------------------------
        out << " 0\n"
                "SECTION\n"
                " 2\n"
                "ENTITIES\n";
        //--------------------------------------------------------------------
        for(segment_vector_index = 0; segment_vector_index < number_of_segments(); segment_vector_index++)
        {
            if(segments[segment_vector_index].number_of_lit_vertices() == 1)
            {
                if(    (   segments[segment_vector_index].first_lit_anchor().position()
                         | segments[segment_vector_index].first_lit_vertex().position()
                       )
                    <= p_space->insignificant_distance
                  )
                {
                    out << " 0\n"
                           "POINT\n"
                           " 62\n  ";
                    out << p_space->
                           palette_picker(LASERBOY_DXF)
                           .best_match(segments[segment_vector_index].first_lit_vertex());
                    out << "\n 10\n";
                    out << segments[segment_vector_index].first_lit_anchor().x;
                    out << "\n 20\n";
                    out << segments[segment_vector_index].first_lit_anchor().y;
                    out << "\n 30\n";
                    out << segments[segment_vector_index].first_lit_anchor().z;
                    out << "\n";
                }
                else
                {
                    out << " 0\n"
                           "LINE\n"
                           " 62\n  ";
                    out << p_space->
                           palette_picker(LASERBOY_DXF)
                           .best_match(segments[segment_vector_index].first_lit_vertex());
                    out << "\n 10\n";
                    out << segments[segment_vector_index].first_lit_anchor().x;
                    out << "\n 20\n";
                    out << segments[segment_vector_index].first_lit_anchor().y;
                    out << "\n 30\n";
                    out << segments[segment_vector_index].first_lit_anchor().z;
                    out << "\n 11\n";
                    out << segments[segment_vector_index].first_lit_vertex().x;
                    out << "\n 21\n";
                    out << segments[segment_vector_index].first_lit_vertex().y;
                    out << "\n 31\n";
                    out << segments[segment_vector_index].first_lit_vertex().z;
                    out << "\n";
                }
            }
            //----------------------------------------------------------------
            else
            {
                out << " 0\n"
                        "POLYLINE\n"
                        " 70\n";
                out << segments[segment_vector_index].is_closed_polygon();
                out << "\n 10\n"
                       "0.0\n"
                       " 20\n"
                       "0.0\n"
                       " 30\n"
                       "0.0\n"
                       //---------
                       " 0\n"
                       "VERTEX\n"
                       " 62\n";
                out << p_space->
                       palette_picker(LASERBOY_DXF)
                       .best_match(segments[segment_vector_index].at(1));
                out << "\n 10\n";
                out << segments[segment_vector_index].first_lit_anchor().x;
                out << "\n 20\n";
                out << segments[segment_vector_index].first_lit_anchor().y;
                out << "\n 30\n";
                out << segments[segment_vector_index].first_lit_anchor().z;
                out << "\n";

                for(vertex_index = 1; vertex_index < segments[segment_vector_index].number_of_vertices() - 1; vertex_index++)
                {
                    out << " 0\n"
                        << "VERTEX\n"
                        << " 62\n";
                    out << p_space->
                           palette_picker(LASERBOY_DXF)
                           .best_match(segments[segment_vector_index].at(vertex_index + 1));
                    out << "\n 10\n";
                    out << segments[segment_vector_index].at(vertex_index).x;
                    out << "\n 20\n";
                    out << segments[segment_vector_index].at(vertex_index).y;
                    out << "\n 30\n";
                    out << segments[segment_vector_index].at(vertex_index).z;
                    out << "\n";
                }
                out << " 0\n"
                       "VERTEX\n"
                       " 62\n"
                       "0\n"
                       " 10\n";
                out << segments[segment_vector_index].at(vertex_index).x;
                out << "\n 20\n";
                out << segments[segment_vector_index].at(vertex_index).y;
                out << "\n 30\n";
                out << segments[segment_vector_index].at(vertex_index).z;
                out << "\n";
            }
        }
        //--------------------------------------------------------------------
        out << " 0\n"
               "ENDSEC\n"
               " 0\n"
               "EOF\n";
        //--------------------------------------------------------------------
        out.close();
        return true;
    }
    return false;
}

//############################################################################
bool LaserBoy_frame::save_as_bmp(string file) const
{
    if(number_of_vertices() > 1)
    {
        char   file_name[80];
        struct LaserBoy_bmp bmp_out = {0};

        if(palette_index == LASERBOY_TRUE_COLOR)
            bmp_init(&bmp_out, p_space->output_bmp_size, p_space->output_bmp_size, 24);
        else
        {
            bmp_init(&bmp_out, p_space->output_bmp_size, p_space->output_bmp_size,  8);
            p_space->palette_picker(palette_index).to_bmp_palette(&bmp_out);
        }
        bmp_clear(&bmp_out, bmp_out.black);

        u_short              i;
        LaserBoy_3D_double   _0,
                             _1;
        int                  bmp_center  = bmp_out.yres / 2;
        double               scale       = bmp_out.yres / 65536.0;
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).is_color())
            {
                _0 = at(i - 1);
                _1 = at(i    );
                //------------------------------------------------------------

                if(palette_index == LASERBOY_TRUE_COLOR)
                    bmp_put_line_fast(   &bmp_out
                                       ,  (int)(_0.x * scale) + bmp_center
                                       , ((int)(_0.y * scale) + bmp_center)
                                       ,  (int)(_1.x * scale) + bmp_center
                                       , ((int)(_1.y * scale) + bmp_center)
                                       , bmp_out.make_rgb(at(i).r, at(i).g, at(i).b)
                                     );
                else
                    bmp_put_line_fast(   &bmp_out
                                       ,  (int)(_0.x * scale) + bmp_center
                                       , ((int)(_0.y * scale) + bmp_center)
                                       ,  (int)(_1.x * scale) + bmp_center
                                       , ((int)(_1.y * scale) + bmp_center)
                                       , at(i).c
                                     );
            }
        }
        sprintf(file_name, "%s", file.c_str());
        bmp_save_as(&bmp_out, file_name);
        bmp_free(&bmp_out);
        return true;
    }
    return false;
}

//############################################################################
LaserBoy_frame blank_frame(LaserBoy_space* ps)
{
    LaserBoy_frame frame(ps);
    frame = blank_segment(ps);
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
LaserBoy_frame NULL_frame(LaserBoy_space* ps)
{
    LaserBoy_frame frame(ps);
    frame = NULL_segment(ps);
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
